def clusters():
    cl = Clusters(self.__pc_df.loc[self.training_ids]) 
    accuracies = [self.__create_models(n_models, data = self.__pc_df.loc[self.training_ids])]
    pbar.update(1)

    for x in range(2, 6):
        clusters = cl.get_cluster_labels(x)
        for i in np.unique(clusters):
            data = self.__pc_df.loc[self.training_ids].iloc[np.where(clusters == i)]
            accuracies.append(self.__create_models(n_models, data))
        pbar.update(1)
    self.__models = self.__models[np.argmax(accuracies)]

    return self.__predict()

def __get_extensive_cluster_df(self) -> pd.DataFrame:
        groups = []
        columns = len(self.__pc_df.columns) * 2
        for group in ["Labels", "Artists", "Companies"]:
            ids = self.__df[group].str.split(";")
            exploded = ids.explode()    
            occurrences = exploded.value_counts()        
            tqdm.pandas(desc = "Finding cluster centers", total = len(exploded.unique()))
            cluster_data = np.concatenate(exploded.to_frame().groupby(group).progress_apply(self.__lookup).values).ravel()
            cluster_data = pd.DataFrame(cluster_data.reshape(int(len(cluster_data) / columns), columns), index = exploded.unique()).fillna(0).apply(pd.to_numeric, downcast="float").round(3)
            cluster_data = cluster_data.multiply(occurrences, axis = 0)

            tqdm.pandas(desc = "Apply cluster centers to releases", total = len(ids.index))
            a = ids.groupby(level=0).progress_apply(lambda x: cluster_data.loc[x.values[0]].sum() / occurrences[x.values[0]].sum()).unstack() # weighted average
            groups.append(a)

        return pd.concat(groups, axis = 1)

   def __lookup(self, x):
        return [self.__pc_df.loc[x.index].mean()]


def calculate_most_similar_group(self, group: str) -> pd.Series:
        groups = {"Release": 0, "Artists": 1, "Labels": 2, "Companies": 3}
        group_size = len(self.__pc_df.columns) / 4
        start = groups[group] * group_size
        end = start + group_size
        mean_group = self.__pc_df.loc[self.__training_ids,start:end].mean()

        group_data = self.__df[group].drop_duplicates().to_frame()
        group_data.columns = ["IDs"]
        group_data["%s Similarity" % group] = 1 / np.array([np.linalg.norm(mean_group - self.__pc_df.loc[i,start:end]) for i in self.__df[group].drop_duplicates().index])
        return group_data.sort_values("%s Similarity" % group, ascending = False)



    def probability_per_release(self, release_id: int) -> pd.Series:
        return self.__predictions.loc[release_id]


import pandas as pd
import numpy as np
import pickle

class Predictions:
    """
    This class allows for post-processing predictions made by the algorithm.
    It allows for saving and loading previous predictions. Releases can be 
    filtered by unseen groups, like Artists and/or Labels.
    """
    def __init__(self, df: pd.DataFrame = None, pc_df: pd.DataFrame = None, 
                training_ids: pd.Series = None, predictions: pd.DataFrame = None):
        self.__df = df
        self.__pc_df = pc_df
        self.__training_ids = training_ids
        self.__predictions = predictions

    def get_top_n(self, n: int=10):
        """
        Returns the N Releases with the highest probability
        """
        return self.__predictions.iloc[:n]
    

    def filter_group(self, group: str) -> pd.Series:
        """
        This method filters out predictions in groups alread present
        in the Wantlist / Collection. This feature can be used to discover
        similar releases on unseen Labels, Artists, or Companies.
        """
        if group not in self.__df.columns:
            raise Exception("""Group not found. Must be one of the following:
            Artists, Labels, Companies""")

        allowed_indexes = self.__df.index[~self.__df[group].isin(self.__df.loc[self.__training_ids, group])]
        mask = np.where(self.__predictions.index.isin(allowed_indexes))
        return self.__predictions.iloc[mask]

    
        

    def load(self, output: str) -> None:
        """
        Loads previously saved predictions.
        """
        with open(output, 'rb') as file:
            tmp_dict = pickle.load(file)     

        self.__dict__.update(tmp_dict) 


    def save(self, output: str) -> None:
        """
        Save the current predictions made.
        """
        with open(output, 'wb') as file:
            pickle.dump(self.__dict__, file, 2)


    def __repr__(self) -> str:
        """
        Prints the top 10 most similar releases as default.
        """
        return "Top 10 most similar Releases:\n{0}".format(self.__predictions.head(10))


from sklearn.cluster import KMeans
from sklearn.metrics import silhouette_score
import logging
import numpy as np
from collections import Counter


class Clusters:
    """

    """
    def __init__(self, training_data) -> None:
        self.__data = training_data
        self.__max_num_of_clust = 5
        self.__minimum = int(len(self.__data.index) / self.__max_num_of_clust)

    def get_cluster_labels(self, n: int):
        """

        """
        clusterer = KMeans(n_clusters=n)
        labels = clusterer.fit_predict(self.__data)
        to_keep = [k for k, c in Counter(labels).items() if c >= self.__minimum]
        return np.array([l if l in to_keep else -1 for l in labels])
    
    def optimal_clustering(self):
        results = {}
        all_labels = {}
        for i in range(2, self.__max_num_of_clust):
            silhouettes = []
            labels = []
            for _ in range(3):
                clusterer = KMeans(n_clusters=i)
                cluster_labels = clusterer.fit_predict(self.__data)
                labels.append(cluster_labels)
                silhouettes.append(silhouette_score(X = self.__data, labels = cluster_labels))
            results[i] = max(silhouettes)
            all_labels[i] = labels[np.argmax(silhouettes)]
            

        logging.debug(results)
        logging.debug("Best clustering = %s" % max(results, key = results.get))
        labels = all_labels.get(max(results, key = results.get))
        to_keep = [k for k, c in Counter(labels).items() if c >= self.__minimum]
        logging.debug(to_keep)
        return np.array([l if l in to_keep else -1 for l in labels])


    def sklearn_approach(self):
        from sklearn.cluster import KMeans
        from sklearn.metrics import silhouette_samples, silhouette_score

        import matplotlib.pyplot as plt
        import matplotlib.cm as cm
        import numpy as np

        X = self.__data
        range_n_clusters = [2, 3, 4, 5, 6]

        for n_clusters in range_n_clusters:
            # Create a subplot with 1 row and 2 columns
            fig, (ax1, ax2) = plt.subplots(1, 2)
            fig.set_size_inches(18, 7)

            # The 1st subplot is the silhouette plot
            # The silhouette coefficient can range from -1, 1 but in this example all
            # lie within [-0.1, 1]
            ax1.set_xlim([-0.1, 1])
            # The (n_clusters+1)*10 is for inserting blank space between silhouette
            # plots of individual clusters, to demarcate them clearly.
            ax1.set_ylim([0, len(X) + (n_clusters + 1) * 10])

            # Initialize the clusterer with n_clusters value and a random generator
            # seed of 10 for reproducibility.
            clusterer = KMeans(n_clusters=n_clusters, random_state=10)
            cluster_labels = clusterer.fit_predict(X)

            # The silhouette_score gives the average value for all the samples.
            # This gives a perspective into the density and separation of the formed
            # clusters
            silhouette_avg = silhouette_score(X, cluster_labels, metric = "mahalanobis")
            print("For n_clusters =", n_clusters,
                "The average silhouette_score is :", silhouette_avg)

            # Compute the silhouette scores for each sample
            sample_silhouette_values = silhouette_samples(X, cluster_labels, metric = "mahalanobis")

            y_lower = 10
            for i in range(n_clusters):
                # Aggregate the silhouette scores for samples belonging to
                # cluster i, and sort them
                ith_cluster_silhouette_values = \
                    sample_silhouette_values[cluster_labels == i]

                ith_cluster_silhouette_values.sort()

                size_cluster_i = ith_cluster_silhouette_values.shape[0]
                y_upper = y_lower + size_cluster_i

                color = cm.nipy_spectral(float(i) / n_clusters)
                ax1.fill_betweenx(np.arange(y_lower, y_upper),
                                0, ith_cluster_silhouette_values,
                                facecolor=color, edgecolor=color, alpha=0.7)

                # Label the silhouette plots with their cluster numbers at the middle
                ax1.text(-0.05, y_lower + 0.5 * size_cluster_i, str(i))

                # Compute the new y_lower for next plot
                y_lower = y_upper + 10  # 10 for the 0 samples

            ax1.set_title("The silhouette plot for the various clusters.")
            ax1.set_xlabel("The silhouette coefficient values")
            ax1.set_ylabel("Cluster label")

            # The vertical line for average silhouette score of all the values
            ax1.axvline(x=silhouette_avg, color="red", linestyle="--")

            ax1.set_yticks([])  # Clear the yaxis labels / ticks
            ax1.set_xticks([-0.1, 0, 0.2, 0.4, 0.6, 0.8, 1])

            # 2nd Plot showing the actual clusters formed
            colors = cm.nipy_spectral(cluster_labels.astype(float) / n_clusters)
            ax2.scatter(X.iloc[:, 0], X.iloc[:, 1], marker='.', s=30, lw=0, alpha=0.7,
                        c=colors, edgecolor='k')

            # Labeling the clusters
            centers = clusterer.cluster_centers_
            # Draw white circles at cluster centers
            ax2.scatter(centers[:, 0], centers[:, 1], marker='o',
                        c="white", alpha=1, s=200, edgecolor='k')

            for i, c in enumerate(centers):
                ax2.scatter(c[0], c[1], marker='$%d$' % i, alpha=1,
                            s=50, edgecolor='k')

            ax2.set_title("The visualization of the clustered data.")
            ax2.set_xlabel("Feature space for the 1st feature")
            ax2.set_ylabel("Feature space for the 2nd feature")

            plt.suptitle(("Silhouette analysis for KMeans clustering on sample data "
                        "with n_clusters = %d" % n_clusters),
                        fontsize=14, fontweight='bold')

        plt.show()


            




from typing import Union
import warnings
import pandas as pd
import numpy as np
import discogs_client
import math
import logging
from pandas._config.config import describe_option
from sklearn.decomposition import PCA
from sklearn.model_selection import train_test_split, cross_validate
from sklearn.ensemble import RandomForestClassifier
from sklearn.metrics import accuracy_score
from tqdm import tqdm
from datetime import datetime
warnings.simplefilter(action='ignore', category=FutureWarning)
from .DiscogsPredictions import Predictions
from .DiscogsClusters import Clusters
import sys

class Learner:
    """
    This class constructs models from extracted Discogs data. It utilizes the official Discogs client to obtain 
    Wantlist and/or Collection identifiers. 
    """
    def __init__(self, db_path: str = "", test: bool = False, token: str = "",
                 use_wantlist: bool = True, use_collection: bool = True, debug: bool = False):

        if not use_wantlist and not use_collection:
            raise Exception("Cannot construct a model without a collection or wantlist.")
        elif token == "":
            raise Exception("""annot construct a model without a personal acces token.
                            See https://www.discogs.com/settings/developers to obtain a token.""")
        
        if debug:
            logging.basicConfig(format='%(levelname)s:%(message)s', level=logging.DEBUG)
        else:
            logging.basicConfig(format='%(levelname)s:%(message)s', level=logging.INFO)
        logging.info("Reading Discogs DB")
        self.training_df = None
        self.__token = token
        self.__pc_df = None
        self.__releases = None
        self.__models = []
        self.__use_wantlist = use_wantlist
        self.__use_collection = use_collection
        self.__df = pd.read_csv(db_path, sep = "\t")
        if test:
            self.__df = self.__df.head(100000)

        self.__corrections()

        
    def __corrections(self) -> None:
        """
        Performs corrections before PCA. Includes scaling, but not centering
        """
        self.__df.ReleaseID = self.__df.ReleaseID.astype(int)
        self.__df.Tracks = self.__df.Tracks.astype(int)
        self.__df = self.__df.replace({"Year": {0: self.__df.Year.median()}, "Tracks": {0: self.__df.Tracks.median()}})
        self.__df.Tracks = self.__df.Tracks.clip(lower = 1, upper = 10)
        self.__df.Year = self.__df.Year.clip(lower = 1900, upper = datetime.today().year)

        self.__df = self.__df.fillna("0")
        self.__df.set_index("ReleaseID", inplace = True)
        self.__df.Year = self.__df.Year / max(self.__df.Year)
        self.__df.Tracks = self.__df.Tracks / max(self.__df.Tracks)
        self.__releases = self.__df.index
        
        self.__add_group_data("Labels")
        self.__add_group_data("Artists")
        self.__add_group_data("Companies")
        self.__df.sort_index(inplace = True)
        

    def __add_group_data(self, group: str) -> None:
        """
        Adds data about groups that should be taken into account when
        creating a PCA. Includes the size of the group and the order of
        releases. 
        """
        self.__df.sort_values([group, "Year"], inplace = True)
        sizes = self.__df[[group, "Year"]].groupby(group).size()
        self.__df["%s_size" % group] = sizes.loc[self.__df[group]].values
        self.__df["%s_sort" % group] = np.concatenate([np.arange(1, x + 1) for x in sizes]) 

        

        self.__df["%s_sort" % group] = self.__df["%s_sort" % group] / max(self.__df["%s_sort" % group])
        self.__df["%s_size" % group] = self.__df["%s_size" % group] / max(self.__df["%s_size" % group])


    def __ids_through_api(self) -> pd.Series:
        """
        Retrieves Release IDs from the wantlsit and/or collection 
        using the Discogs client
        """
        d = discogs_client.Client('Discogs Learner/0.1', user_token=self.__token)
        me = d.identity()
        collection = []
        wantlist = []
        if self.__use_collection:
            collection = [release.id for release in me.collection_folders[0].releases]
        
        if self.__use_wantlist:
            wantlist = [release.id for release in me.wantlist]
        return pd.Series(wantlist + collection)

        
    def __set_training_set(self) -> None:
        """
        Creates a subset of the database using release IDs
        """
        logging.info("Fetching IDs from Discogs")
        self.training_ids = self.__ids_through_api()
        self.training_df = self.__df.loc[self.training_ids.iloc[np.where(
            self.training_ids.isin(self.__df.index))]].drop_duplicates()
        self.training_ids = self.training_df.index
        logging.debug("Identifiers found in discogs: %s" % len(self.training_ids))

    def __format_df(self, x: pd.DataFrame) -> pd.DataFrame:
        """
        Formats a subset of the database for use in PCA.
        """
        styles = pd.get_dummies(x.Styles.str.split(";").explode()).groupby(level=0).sum()
        countries = pd.get_dummies(x.Country.explode())
        formats = pd.get_dummies(x.Formats.explode())
        df = pd.concat([styles, countries, formats, x.Year, x.Tracks], axis = 1) 
        df.columns = np.arange(len(df.columns))
        return df

    def __get_pca(self) -> tuple([PCA, list]):
        """
        This method creates a PCA transformation from the collection / wantlist.
        This allows to quantify important factors for the models.
        """
        df = self.__format_df(self.training_df)
        components = int(round(math.sqrt(len(df.columns)))) if len(df.columns) <= 15 else 15
        logging.debug("Number of PCs = %s" % components)
        pca = PCA(n_components=components)
        pca.fit(df)
        logging.debug("Variance explained: %s " % pca.explained_variance_ratio_)
        return pca, df.columns
    
    def __reduce_database(self) -> None:
        """
        This method reduces the database by only selecting Styles
        that appear in the wantlist / collection. This due to Releases
        only being interesting if their styles match.  
        """
        styles_per_release = self.__df.Styles.str.split(";").explode()
        allowed = self.training_df.Styles.str.split(";").explode().value_counts()
        
        compare = pd.concat([
            styles_per_release.iloc[np.where(styles_per_release.isin(allowed.index))].index.value_counts(),
            styles_per_release.index.value_counts()
        ], axis = 1)
        
        self.__df.set_index(self.__releases, inplace = True)
        self.__df = self.__df.iloc[np.where(compare.iloc[:,0] == compare.iloc[:,1])]
        self.__releases = self.__df.index

        self.training_df = self.__df.loc[self.training_ids[np.where(self.training_ids.isin(self.__df.index))]].drop_duplicates()
        self.training_ids = self.training_df.index
        logging.debug("Identifiers found after reduction: %s" % len(self.training_ids))

        
    def __transform_to_pc_space(self, x, pca: PCA, columns: list) -> pd.DataFrame:
        """
        Transforms a dataframe to PCA space. This is used for the training
        data, but also for the discogs database.
        """
        sub = self.__format_df(x)
        for c in columns:
            if c not in sub.columns:
                sub[c] = 0
        return pd.DataFrame(pca.transform(sub[columns]), index = sub.index)


    def __get_cluster_df(self) -> pd.DataFrame:
        """
        This method returns the mean and variance of a given group and maps
        those to individual releases. 
        """
        groups = []
        columns = ["Artists", "Labels", "Companies"]
        for i, group in enumerate(columns):
            tqdm.pandas(desc = "Creating %s map (%s/%s)" % (group, i + 1, len(columns)), total = len(self.__df[group].unique()))
            cluster_data = self.__pc_df.groupby(self.__df[group]).progress_apply(lambda x: x.mean()).apply(pd.to_numeric, downcast="float").round(3)
            cluster_data.columns = np.arange(0, len(cluster_data.columns)) 
            means = cluster_data.loc[self.__df[group]].set_index(self.__pc_df.index)
            groups.append(pd.concat([means, self.__pc_df.subtract(means) ** 2], axis = 1))
        return pd.concat(groups, axis = 1)

    def __adjust_pc_df(self, pca: PCA, columns: list) -> None:
        """
        This method maps the Discogs database to PCA space and adds 
        the mean and variance of groups to the PCA dataframe. To save RAM,
        this is done in batches.
        """
        self.__df.reset_index(inplace = True) 
        tqdm.pandas(desc = "Mapping database to PCA space", total = len(self.__df.index) // 5000)
        self.__pc_df = self.__df.groupby(self.__df.index // 5000).progress_apply(
            lambda x: self.__transform_to_pc_space(x, pca, columns)
        ).reset_index(drop=True).apply(pd.to_numeric, downcast="float").round(3)
        self.__df.set_index("ReleaseID", inplace = True)
        self.__df = self.__df[["Labels", "Artists", "Companies"]]
        self.__pc_df.set_index(self.__df.index, inplace = True)
        self.__pc_df = pd.concat([self.__pc_df, self.__get_cluster_df()], axis = 1)
        self.__pc_df.columns = np.arange(0, len(self.__pc_df.columns))
        logging.debug("Final PC DataFrame:\n%s" % self.__pc_df.head())


    def __create_train_test(self, X1: pd.DataFrame) -> list:
        """
        This method creates a train-test split using 
        the training identifiers determined from the 
        collection / wantlist.
        """
        
        X1["Pred"] = 1
        X2 = self.__pc_df.sample(len(X1.index))
        X2["Pred"] = 0
        
        X = pd.concat([X1, X2])
        labels = X.Pred
        X.drop("Pred", inplace = True, axis = 1)
        return train_test_split(X, labels, test_size=0.3)
        
    def __create_models(self, n_models: int, data: pd.DataFrame) -> None:
        """
        Creates N models. Uses a Random Forest classifier with 
        5-fold cross-validation to prevent overfitting.
        """
        scores = []
        models = []
        with tqdm(desc = "Creating models", total=n_models) as pbar:
            for _ in range(n_models):
                X_train, X_test, y_train, y_test = self.__create_train_test(data)
                clf = RandomForestClassifier(n_estimators = 100)
                cv_results = cross_validate(clf, X_train, y_train, 
                                            cv=5, scoring = "accuracy", 
                                            return_estimator = True)
                model = cv_results["estimator"][np.argmax(cv_results["test_score"])]
                scores.append(accuracy_score(y_test, model.predict(X_test)))
                self.__models.append(model)
                pbar.update(1)
        score = np.mean(scores)
        logging.debug("Model accuracy: %s" % score)
        return score

    def learn_and_predict(self, n_models: int=10) -> list:
        """
        The main method to be called for training models and 
        predicting Releases. Uses 10 models by default, should
        use less than 5 models.
        Returns a sorted Pandas Series object with Release IDs 
        as index and probabilities as values. 
        """
        self.__set_training_set()
        self.__reduce_database()
        pca, columns = self.__get_pca()
        self.__adjust_pc_df(pca, columns)
        self.__create_models(n_models, data = self.__pc_df.loc[self.training_ids])
        return self.__predict()      

    def __predict(self) -> Predictions:
        """
        Predict similarity of releases to the trained models. 
        Requires 'start_learning' to be called first. 
        Returns a sorted Pandas Series object with Release IDs 
        as index and probabilities as values.  
        """
        total = []
        with tqdm(desc = "Predicting Releases", total=len(self.__models)) as pbar:
            for model in self.__models:
                total.append(pd.DataFrame(model.predict_proba(self.__pc_df), index=self.__pc_df.index)[1])
                pbar.update(1)
            preds = pd.concat(total, axis = 1).prod(axis=1).sort_values(ascending=False)
            preds = preds.iloc[np.where(preds.index.isin(self.training_ids) == False)].to_frame()
        return Predictions(
            df = self.__df,
            pc_df = self.__pc_df,
            training_ids= self.training_ids,
            predictions=preds
        )

    
    
    



import json
import requests
import re
import pandas as pd
from .DiscogsChunker import _Chunker as Chunker
from datetime import datetime
from urllib.request import urlopen, Request
from tqdm import tqdm
import xml.etree.cElementTree as ET
import sys
from pathlib import Path
import os

class Extracter:
    """
    This class extracts data from the monthly Discogs datadumps.
    By default it selects the recent Releases, but a custom URL
    can be given if needed.
    """
    def __init__(self, genre: str="Electronic", url: str = None):
        today = datetime.today()
        self.__url = url
        if url is None:
            self.__url = "https://discogs-data.s3-us-west-2.amazonaws.com/data/%d/discogs_%d%02d01_releases.xml.gz" % (today.year, today.year, today.month)     

        self.__main = []
        self.__genre = genre
        self.__mapping_styles = {}
        self.__mapping_formats = {}
        self.__mapping_countries = {}
        self.__n_releases = self.__get_n_releases()
    
    def __get_n_releases(self) -> int:
        """
        Retrieves the number of releases in the database to use as indicator.
        """
        url = "https://api.discogs.com/"
        response = json.loads(requests.get(url).text)
        return int(response["statistics"]["releases"])
            
    def extract(self, output: str) -> None:
        """
        Main method for extracting Discogs data. Uses the Chunker class to
        cut up the XML and save RAM. Needs an output file for the data
        to be stored. Has to be run only once per month to keep updated.
        """
        if output is None:
            raise Exception("No output file given.")
        
        os.makedirs(os.path.dirname(output), exist_ok=True)

        a = open(output, "w")
        a.write("\t".join(["ReleaseID", "Labels", "Formats", "Year", "Country","Artists", 
                               "Styles", "Tracks", "Companies"]) + "\n")
        a.close()

        with urlopen(Request(self.__url, headers={"Accept-Encoding": "gzip"})) as response:
            c = Chunker(response, chunk_by = "release")
            with tqdm(desc = "Parsing releases from Discogs", total=self.__n_releases) as pbar:
                for i, chunk in enumerate(c.chunk()):
                    self.__extract_chunk(chunk)
                    if i % 100000 == 0:
                        self.__write(pd.DataFrame(self.__main), output)
                    pbar.update(1)
            self.__write(pd.DataFrame(self.__main), output)

    def __get_mapping(self, mapping: dict, x: str) -> int:
        """
        Converts a new value to a integer, saving storage space.
        """
        return mapping.setdefault(x, len(mapping))
        
    def __extract_chunk(self, chunk: ET) -> None:
        """
        Finds values in a Release and adds them to a list. 
        """
        if chunk.findtext(".//genre", "Unknown") == self.__genre:
            id = chunk.attrib["id"]
            released = int(re.sub(r'[^0-9]+', '0', chunk.findtext("released", "0"))[:4])
            labels = ";".join({re.sub(r'[^0-9]+', "0", l.attrib.get("id", "0")) for l in chunk.findall(".//label")})
            country = self.__get_mapping(self.__mapping_countries, chunk.findtext("country", "Unknown"))
            styles = ";".join({str(self.__get_mapping(self.__mapping_styles, x.text)) for x in chunk.findall(".//style")})
            artists = ";".join({x[0].text for x in chunk.findall(".//artist")})
            tracks = len(chunk.findall("tracklist//track//title"))
            format = self.__get_mapping(self.__mapping_formats, chunk.find("formats//format").attrib["name"])
            company = ";".join({x[0].text for x in chunk.findall("companies//company")}) 
            self.__main.append((id, labels, format, released, country, artists, styles, tracks, company))

    def __write(self, df: pd.DataFrame, target: str) -> None:
        """
        Writes processed chunks to the given file.
        """
        df.to_csv(target, mode = "a", header = False, sep = "\t", index = False)
        self.__main = []


from setuptools import setup

from os import path
this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

version = long_description.split("Version: ")[1][0:3]

setup(
   name='discogslearner',
   version=version,
   license = "GPL-3.0",
   description='Machine Learning module for Discogs',
   author='Pascal Maas',
   author_email='p.maas92@gmail.com',
   packages=['discogslearner'],
   install_requires=['pandas', 'tqdm', 'numpy', "sklearn", "discogs_client"],
   url = "https://github.com/Pascallio",
   download_url = "https://github.com/Pascallio/DiscogsLearner/archive/refs/tags/v0.2.tar.gz",
   keywords = ["Discogs", "Machine Learning"],
   long_description=long_description,
   long_description_content_type='text/markdown'
)


from typing import Any, Iterator
import xml.etree.cElementTree as ET
from gzip import GzipFile
import sys

class _Chunker:
    """
    This class cuts up large XML files into smaller chunks
    given the 'chunk_by' argument. The source needs to be a 
    URL to a gzipped xml file. This base class is made for
    the Discogs datadump releases XML, and even though it is 
    generic, it might not work for differently formatted XMLs.
    """
    def __init__(self, source: Any, chunk_by: str):
        self.__source = source
        self.__chunk_by = "</{}>".format(chunk_by)

    def chunk(self) -> Iterator:
        """
        This is the main method of the DiscogsChunker. It unpacks
        the Gzip file and returns a generator for a chunked XML.
        It returns an ElementTree object.
        """
        with GzipFile(fileobj=self.__source) as f:
            f.readline()
            data = ""
            for line in f:
                s = line.decode('UTF-8')
                data += s
                if self.__chunk_by in s:
                    yield ET.fromstring(data)
                    data = ""